use std::collections::BTreeMap;

use snafu::{ResultExt, Snafu};

/// The types and directives in a VijiN project.
pub struct TypeSystem {
    pub types: BTreeMap<String, Type>,
    pub directives: BTreeMap<String, Directive>,
}

/// The definition of a standard directive.
///
/// Note that this type is an input to the parser, not an output.
pub struct Directive {
    /// The provider's name for the directive.
    name: String,
    /// Description of the directive.
    doc: String,
    /// Fixed prefix term.
    prefix: String,
    /// The type of the prefix parameter, if there is one.
    prefix_param: Option<Param>,
    /// Fixed infix term.
    /// Can only be present if the prefix parameter exists.
    infix: Option<String>,
    /// The type of the infix parameter.
    /// Can only be present if the infix term exists.
    infix_param: Option<Param>,
    /// Fixed suffix term.
    suffix: Option<String>,
    /// The type of the suffix parameter.
    /// Can only be present if the suffix exists.
    suffix_param: Option<Param>,
    /// The type of the elements of the list parameter, if supported.
    list_param: Option<Param>,
    // TODO(zeph): macro expansion?
}

/// Directive parameter.
pub struct Param {
    /// Name to which the compiler maps the argument.
    name: String,
    /// Description of the parameter's meaning in the context of the directive.
    doc: Option<String>,
    /// Type required for the argument.
    typ: Type,
    /// Nuance parameter for list parameters.
    nuance: Option<Box<Param>>,
}

/// Types expected by directive arguments.
#[derive(Clone)]
pub enum Type {
    /// Arbitrary text.
    Text,
    /// Speech text, possibly containing emphasis, shout, or whisper markers.
    Speech,
    /// Character in the current scene.
    Character,
    /// Duration.
    Duration,
    /// Asset requiring particular tags.
    Asset { tags: Vec<String> },
    /// Enumeration of literal options.
    Enum(Vec<String>),
}

// Serialized representation of directives.
mod serial {
    use serde::Deserialize;

    #[derive(Deserialize)]
    #[serde(rename_all = "snake_case", deny_unknown_fields)]
    pub struct Document {
        pub types: Vec<Type>,
        pub directives: Vec<Directive>,
    }

    #[derive(Deserialize)]
    #[serde(rename_all = "snake_case", deny_unknown_fields, tag = "kind")]
    pub enum TypeKind {
        Text,
        Speech,
        Character,
        Duration,
        Asset {
            tags: Vec<String>,
        },
        Enum {
            #[serde(rename = "enum")]
            enm: Vec<String>,
        },
    }

    #[derive(Deserialize)]
    #[serde(rename_all = "snake_case", deny_unknown_fields)]
    pub struct Type {
        pub name: String,
        pub doc: String,
        pub kind: TypeKind,
    }

    #[derive(Deserialize)]
    #[serde(rename_all = "snake_case", deny_unknown_fields)]
    pub struct Param {
        pub name: String,
        pub doc: Option<String>,
        #[serde(rename = "type")]
        pub typ: String,
        pub nuance: Option<Box<Param>>,
    }

    #[derive(Deserialize)]
    #[serde(rename_all = "snake_case", deny_unknown_fields)]
    pub struct Directive {
        pub name: String,
        pub doc: Option<String>,
        pub prefix: String,
        pub prefix_param: Option<Param>,
        pub infix: Option<String>,
        pub infix_param: Option<Param>,
        pub suffix: Option<String>,
        pub suffix_param: Option<Param>,
        pub list_param: Option<Param>,
    }
}

#[derive(Debug, Snafu)]
pub enum Error {
    #[snafu(display("invalid directive definition file: {source}"))]
    Decode { source: serde_json::Error },

    #[snafu(display("no type {name} of {position} parameter for {directive}"))]
    ParamType {
        directive: String,
        position: &'static str,
        name: String,
    },

    #[snafu(display("{directive} has {position} parameter with no {position}"))]
    MissingSyntax {
        directive: String,
        position: &'static str,
    },

    #[snafu(display("{directive} has illegal nuance on {position}"))]
    InvalidNuance {
        directive: String,
        position: &'static str,
    },
}

static DIRECTIVE_PREDECLARED: &str = include_str!("directive-predeclared.json");

impl TypeSystem {
    pub fn new() -> TypeSystem {
        let mut ts = TypeSystem {
            types: BTreeMap::new(),
            directives: BTreeMap::new(),
        };
        ts.parse(DIRECTIVE_PREDECLARED).unwrap();
        ts
    }

    pub fn parse(&mut self, src: &str) -> Result<&mut Self, Error> {
        let s: serial::Document = serde_json::from_str(src).context(DecodeSnafu)?;
        for t in s.types {
            let p = match t.kind {
                serial::TypeKind::Text => Type::Text,
                serial::TypeKind::Speech => Type::Speech,
                serial::TypeKind::Character => Type::Character,
                serial::TypeKind::Duration => Type::Duration,
                serial::TypeKind::Asset { tags } => Type::Asset { tags },
                serial::TypeKind::Enum { enm } => Type::Enum(enm),
            };
            self.types.insert(t.name, p);
        }
        for d in s.directives {
            let prefix_param = d
                .prefix_param
                .map(|p| self.param(&d.name, "prefix", p))
                .transpose()?;
            let infix_param = d
                .infix_param
                .map(|p| {
                    if d.suffix.is_none() {
                        return Err(Error::MissingSyntax {
                            directive: d.name.clone(),
                            position: "infix",
                        });
                    }
                    self.param(&d.name, "infix", p)
                })
                .transpose()?;
            let suffix_param = d
                .suffix_param
                .map(|p| {
                    if d.suffix.is_none() {
                        return Err(Error::MissingSyntax {
                            directive: d.name.clone(),
                            position: "suffix",
                        });
                    }
                    self.param(&d.name, "suffix", p)
                })
                .transpose()?;
            let list_param = d
                .list_param
                .map(|p| self.param(&d.name, "list", p))
                .transpose()?;
            let r = Directive {
                name: d.name.clone(),
                doc: d.doc.unwrap_or_default(),
                prefix: d.prefix,
                prefix_param,
                infix: d.infix,
                infix_param,
                suffix: d.suffix,
                suffix_param,
                list_param,
            };
            self.directives.insert(d.name, r);
        }
        Ok(self)
    }

    fn param(
        &self,
        directive: &str,
        position: &'static str,
        p: serial::Param,
    ) -> Result<Param, Error> {
        let t = match self.types.get(&p.typ) {
            None => {
                return Err(Error::ParamType {
                    directive: directive.into(),
                    position,
                    name: p.typ,
                })
            }
            Some(t) => t,
        };
        let nuance = if let Some(n) = p.nuance {
            if position != "list" {
                return Err(Error::InvalidNuance {
                    directive: directive.into(),
                    position,
                });
            }
            Some(Box::new(self.param(directive, "list nuance", *n)?))
        } else {
            None
        };
        Ok(Param {
            name: p.name,
            doc: p.doc,
            typ: t.clone(),
            nuance,
        })
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use rstest::rstest;
}
